/* GPLv2 applies
 * SVN revision: $Revision: 671 $
 * (C) 2006-2014 by folkert@vanheusden.com
 */
char *soundex(const char *in);
BOOL fuzzy_match(const char *haystack, const char *needle, char *bitmap);
